from typing import *
import logging
import numpy as np

try:
    from rpy2.robjects.packages import importr
    import rpy2.robjects as ro
    from rpy2.robjects import numpy2ri

    def convert_r_obj(v: Any, obj_to_obj: bool=True, verbose: bool=True) -> Any:
        """Function with manually specified conversion from a r-object to a python object
        """
        if type(v) == ro.rinterface.RNULLType:
            return None
        elif type(v) == ro.vectors.Matrix:
            return np.array(v)
        elif type(v) == ro.vectors.FloatVector:
            return np.array(v, dtype="float64")
        elif type(v) == ro.vectors.IntVector:
            return np.array(v, dtype="int64")
        elif type(v) == ro.rinterface.RNULLType:
            return None
        elif type(v) == ro.vectors.ListVector:
            try:
                return {v.names[i]: convert_r_obj(v[i], obj_to_obj=obj_to_obj) for i in range(len(v))}
            except TypeError:
                return {i: convert_r_obj(v[i], obj_to_obj=obj_to_obj) for i in range(len(v))}
        elif type(v) == ro.vectors.StrVector:
            if len(v) == 1:
                return str(v[0])
            else:
                try:
                    return {v.names[i]: convert_r_obj(v[i], obj_to_obj=obj_to_obj) for i in range(len(v))}
                except TypeError:
                    return {i: convert_r_obj(v[i], obj_to_obj=obj_to_obj) for i in range(len(v))}
        elif type(v) == ro.vectors.DataFrame:
            from rpy2.robjects import pandas2ri
            return pandas2ri.ri2py(v)
        elif type(v) == ro.methods.RS4:
            if obj_to_obj:
                class RS4Object(object):
                    def __repr__(self) -> str:
                        return f"< RS4Object with attributes: {list(self.__dict__.keys())} >"
                rs4obj = RS4Object()
                for k in tuple(v.slotnames()):
                    setattr(rs4obj, k, convert_r_obj(v.slots[k], obj_to_obj=obj_to_obj))
                return rs4obj
            else:
                return {k: convert_r_obj(v.slots[k]) for k in tuple(v.slotnames())}
        else:
            if type(v) != str:
                if verbose:
                    print(f"not supported yet {type(v)}")
            return v

except:
    pass



from typing import *
import numpy as np
import velocyto as vcy


def jump_next_3p_exon(feature: vcy.Feature) -> vcy.Feature:
    """Jump to the next exon following transcription direction instead of chromosome coordinate

    Arguments
    ---------
    feature: vcy.Feature
        An exonic feature

    Returns
    -------
    vcy.Feature:
        The next 3' exon following transcription direction

    Note
    ----
    It raises IndexError if the Feature is the 3'-most feature in the transcript model
    """
    if feature.transcript_model.chromstrand[-1] == "+":
        ix = feature.exin_no * 2
    else:
        ix = len(feature.transcript_model.list_features) - 2 * (feature.exin_no - 1) - 3
        if ix < 0:
            raise IndexError
    return feature.transcript_model.list_features[ix]


def closest_3prime(segment_match: vcy.SegmentMatch) -> int:
    """Calculate the closest distance walking on the transcript model to the 3'UTR

    Argument
    --------
    segment_match: vcy.SegmentMatch
        the sement from which 5' extremity calculate the distance

    Returns
    -------
    int:
        Distance in base pairs

    Note
    ----
    It skips all the introns but the one where the segment is mapping (if mapping is intronic)

    """
    dist23prime: int = 0
    if segment_match.feature.transcript_model.chromstrand[-1] == "+":
        if segment_match.maps_to_exon:
            curr_exon = segment_match.feature
            to_end_of_exon = curr_exon.end - segment_match.segment[0] + 1
        else:  # maps_to_intron
            curr_intron = segment_match.feature
            to_end_of_exon = curr_intron.end - segment_match.segment[0] + 1
            curr_exon = curr_intron.get_downstream_exon()
            to_end_of_exon += len(curr_exon)

        dist23prime += to_end_of_exon
        while True:
            try:
                curr_exon = jump_next_3p_exon(curr_exon)
                dist23prime += len(curr_exon)
            except IndexError:
                break
    else:  # "-" strand
        if segment_match.maps_to_exon:
            curr_exon = segment_match.feature
            to_end_of_exon = segment_match.segment[-1] - curr_exon.start + 1
        else:  # maps_to_intron
            curr_intron = segment_match.feature
            to_end_of_exon = segment_match.segment[-1] - curr_intron.start + 1
            curr_exon = curr_intron.get_upstream_exon()
            to_end_of_exon += len(curr_exon)

        dist23prime += to_end_of_exon
        while True:
            try:
                curr_exon = jump_next_3p_exon(curr_exon)
                dist23prime += len(curr_exon)
            except IndexError:
                break
    return dist23prime


def spliced_iter(segments_list: List[vcy.SegmentMatch], read_len: int=99) -> Iterable[vcy.SegmentMatch]:
    """Iterates over a list of segment matches grouping spliced segment in new special segment matches
    The goal i s to make the yielded output to be compatible with closest_3prime and further 3' mapping dist pipeline

    Arguments
    ---------
    segments_list: List[vcy.SegmentMatch]
        A list of segment matches objects

    read_len: int, default=99
        The length of a illumin read in the given technology

    Returns
    -------
    Iterable[vcy.SegmentMatch]:
        Either the original segment match or a fake Segment matche that will take the place of a set segment that are spliced

    Note
    ----
    This does not take in consideration all the corner cases it is very difficoult without keeping track of the splicing event
    """
    segments_list = list(segments_list)  # avoid popping way results from the original
    i = 0
    while len(segments_list):
        sm = segments_list.pop(0)
        if sm.is_spliced:
            sm_list = [sm]
            while segments_list[0].is_spliced:
                sm_list.append(segments_list.pop(0))
                if sum([s.segment[1] - s.segment[0] + 1 for s in sm_list]) + segments_list[0] > read_len:
                    break
            if len(segments_list) != 2:
                # just for safety let's ignore those counts otherwise we could make a mess
                continue
                
            if sm_list[0].feature.transcript_model.chromstrand[-1] == "+":
                if sm_list[-1].feature.kind == ord("i"):
                    i += len(sm_list)
                    yield vcy.SegmentMatch(segment=sm_list[0].segment,
                                           feature=sm_list[-1].feature)
                else:  # ord("e")
                    # in this way the distance should be calculated correctly
                    i += len(sm_list)
                    yield vcy.SegmentMatch(segment=(sm_list[-1].feature.start - (sm_list[0].segment[-1] - sm_list[0].segment[0]), -1),
                                           feature=sm_list[-1].feature)
            else:  # "-" strand
                if sm_list[0].feature.kind == ord("i"):
                    i += len(sm_list)
                    yield vcy.SegmentMatch(segment=sm_list[-1].segment,
                                           feature=sm_list[0].feature)
                else:  # ord("e")
                    i += len(sm_list)
                    yield vcy.SegmentMatch(segment=(-1, sm_list[-1].feature.end + (sm_list[0].segment[-1] - sm_list[0].segment[0]), -1),
                                           feature=sm_list[0].feature)
        else:
            i += 1
            yield sm



MATCH_INSIDE = 1
MATCH_OVER5END = 2
MATCH_OVER3END = 4

MIN_FLANK = 5
PATCH_INDELS = 3
SPLIC_INACUR = 6
MIN_POLYT = 8
MAX_USHORT = 2**16 - 1

LOOM_NUMERIC_DTYPE = "uint16"

EXTENSION5_LEN = 0  # basepairs to extend 5' ends of models
EXTENSION3_LEN = 0  # basepairs to extend 3' ends of models

BINSIZE_BP = 100000  # binsize to look up for other transcripts used during interval determination
LONGEST_INTRON_ALLOWED = 1000000
BAM_COMPRESSION = 7

EXON = 1
ALT_EXON = 2  # From another transcript of same gene
OTHER_EXON = 4  # From an unrelated gene
INTRON = 8
ALT_INTRON = 16
OTHER_INTRON = 32
COMPETING_INTRON = ALT_INTRON | OTHER_INTRON
COMPETING_EXON = ALT_EXON | OTHER_EXON

PLACEHOLDER_UMI_LEN = 12  # the length of the placeholder random umi added if -U is set (complexity is 36 per unit length)

CIGAR = {0: "BAM_CMATCH",
         1: "BAM_CINS",
         2: "BAM_CDEL",
         3: "BAM_CREF_SKIP",
         4: "BAM_CSOFT_CLIP",
         5: "BAM_CHARD_CLIP",
         6: "BAM_CPAD",
         7: "BAM_CEQUAL",
         8: "BAM_CDIFF",
         9: "BAM_CBACK"}  # type: dict  # currently hard coded for speed

GEM_codes = {'SI-3A-A1': ['AAACGGCG', 'CCTACCAT', 'GGCGTTTC', 'TTGTAAGA'],
             'SI-3A-A10': ['ACAGCAAC', 'CGCAATTT', 'GAGTTGCG', 'TTTCGCGA'],
             'SI-3A-A11': ['ACCAGTCC', 'CTTTCCTT', 'GGACAGGG', 'TAGGTAAA'],
             'SI-3A-A12': ['ACTACTGT', 'CGGGAACG', 'GACCTCTC', 'TTATGGAA'],
             'SI-3A-A2': ['AGCCCTTT', 'CAAGTCCA', 'GTGAGAAG', 'TCTTAGGC'],
             'SI-3A-A3': ['AAAGCATA', 'CTGCAGCC', 'GCCTTTAT', 'TGTAGCGG'],
             'SI-3A-A4': ['AGAACGCC', 'CATGGCAG', 'GTCTTTGA', 'TCGCAATT'],
             'SI-3A-A5': ['ATTGGGAA', 'CAGTCTGG', 'GGCATACT', 'TCACACTC'],
             'SI-3A-A6': ['ACGGGACT', 'CTTTCGAC', 'GAACATGA', 'TGCATCTG'],
             'SI-3A-A7': ['AGGTCATA', 'CTCATCAT', 'GCTGAGGG', 'TAACGTCC'],
             'SI-3A-A8': ['ATGATACG', 'CCACAGAA', 'GACTGTTC', 'TGTGCCGT'],
             'SI-3A-A9': ['ACAACTTG', 'CTCCAACA', 'GAGTGCGT', 'TGTGTGAC'],
             'SI-3A-B1': ['AGGCTACC', 'CTAGCTGT', 'GCCAACAA', 'TATTGGTG'],
             'SI-3A-B10': ['ACCATTAA', 'CTGGACGT', 'GAACGGTC', 'TGTTCACG'],
             'SI-3A-B11': ['ATGGTCGC', 'CGACATAG', 'GATTCGCT', 'TCCAGATA'],
             'SI-3A-B12': ['ACGCTTGG', 'CGCTACAT', 'GAAAGACA', 'TTTGCGTC'],
             'SI-3A-B2': ['AAGTTGAT', 'CCCACCCA', 'GGTCGAGC', 'TTAGATTG'],
             'SI-3A-B3': ['ATTGGACG', 'CAGCTTAC', 'GGCAAGGA', 'TCATCCTT'],
             'SI-3A-B4': ['AGGGACTG', 'CCTCTAAC', 'GACAGGCT', 'TTATCTGA'],
             'SI-3A-B5': ['ATCGTACT', 'CATCAGTG', 'GGGACTAC', 'TCATGCGA'],
             'SI-3A-B6': ['AACGCGAA', 'CTATTTGG', 'GCGCACCT', 'TGTAGATC'],
             'SI-3A-B7': ['AGGGATGA', 'CTTCTGTT', 'GAATGCAC', 'TCCACACG'],
             'SI-3A-B8': ['ACGTTCAC', 'CAAGGTCT', 'GTTAAGTG', 'TGCCCAGA'],
             'SI-3A-B9': ['AAGCGTGT', 'CTTGACCG', 'GCCACGTA', 'TGATTAAC'],
             'SI-3A-C1': ['AGACTTTC', 'CCGAGGCA', 'GATGCAGT', 'TTCTACAG'],
             'SI-3A-C10': ['ATCTGATC', 'CGTGCTAA', 'GAGAAGGG', 'TCACTCCT'],
             'SI-3A-C11': ['ACCGAACA', 'CGACTCTT', 'GTTTGTGG', 'TAGACGAC'],
             'SI-3A-C12': ['ATCCGGCA', 'CCGTTATG', 'GGTAATGT', 'TAAGCCAC'],
             'SI-3A-C2': ['AATCACTA', 'CCGAGAAC', 'GTAGTGCG', 'TGCTCTGT'],
             'SI-3A-C3': ['ACGTTACA', 'CGTAGGTT', 'GACGACGG', 'TTACCTAC'],
             'SI-3A-C4': ['ACATTGGC', 'CTTAGTCA', 'GAGCCCAT', 'TGCGAATG'],
             'SI-3A-C5': ['ATGCATTC', 'CACTGACT', 'GGTACGGG', 'TCAGTCAA'],
             'SI-3A-C6': ['ACTCAGAC', 'CGCTCAGG', 'GAGGTTTA', 'TTAAGCCT'],
             'SI-3A-C7': ['ACACCGGG', 'CATAATCC', 'GGCGGAAT', 'TTGTTCTA'],
             'SI-3A-C8': ['AGCTCGAG', 'CAGGAAGA', 'GCACGTTT', 'TTTATCCC'],
             'SI-3A-C9': ['AGATCGGT', 'CATCGTCG', 'GTCATATA', 'TCGGACAC'],
             'SI-3A-D1': ['AGCTGCGT', 'CAACCATC', 'GTGGAGCA', 'TCTATTAG'],
             'SI-3A-D10': ['AGATAACA', 'CTTATTTG', 'GCGGGCAT', 'TACCCGGC'],
             'SI-3A-D11': ['ATATGAGA', 'CACCTCAG', 'GCTACTTC', 'TGGGAGCT'],
             'SI-3A-D12': ['AGAAACGT', 'CACTCAAC', 'GCTGTGTA', 'TTGCGTCG'],
             'SI-3A-D2': ['ACATTCCG', 'CTGCGGTA', 'GACACAAT', 'TGTGATGC'],
             'SI-3A-D3': ['ACTTCACT', 'CGAAGTTG', 'GAGCACGC', 'TTCGTGAA'],
             'SI-3A-D4': ['AAATCGTC', 'CTTCGAAT', 'GCGATCGG', 'TGCGATCA'],
             'SI-3A-D5': ['AGACGGAT', 'CCTTTAGA', 'GTCGACTC', 'TAGACTCG'],
             'SI-3A-D6': ['ATGCCAAA', 'CCTTATCG', 'GAAGTCTT', 'TGCAGGGC'],
             'SI-3A-D7': ['AACTTAGA', 'CCGGATCC', 'GGTCGCAT', 'TTAACGTG'],
             'SI-3A-D8': ['AATCTTTG', 'CTCAAGAC', 'GGATGAGT', 'TCGGCCCA'],
             'SI-3A-D9': ['ACCTACTG', 'CAAGGGAC', 'GGGACACA', 'TTTCTTGT'],
             'SI-3A-E1': ['ACGAAAGC', 'CGCCCGTA', 'GTTTGCCT', 'TAAGTTAG'],
             'SI-3A-E10': ['ATTGTTTC', 'CGCAGGAG', 'GCACCAGT', 'TAGTACCA'],
             'SI-3A-E11': ['ATCGCCAT', 'CATAAAGG', 'GGGTTTCC', 'TCACGGTA'],
             'SI-3A-E12': ['ACGCGGAA', 'CGCTATCC', 'GTTGCATG', 'TAAATCGT'],
             'SI-3A-E2': ['AGGCTGGT', 'CACAACTA', 'GTTGGTCC', 'TCATCAAG'],
             'SI-3A-E3': ['AACAAGTC', 'CGGCTCCA', 'GTATGTAT', 'TCTGCAGG'],
             'SI-3A-E4': ['AGCTGACG', 'CCGGTGTC', 'GTAAACAT', 'TATCCTGA'],
             'SI-3A-E5': ['ATCCAAGG', 'CCGTTGAA', 'GGAAGCTC', 'TATGCTCT'],
             'SI-3A-E6': ['ATTGAAAC', 'CAGCCCGA', 'GCCATTTG', 'TGATGGCT'],
             'SI-3A-E7': ['AAGACGTG', 'CCATGTGT', 'GTTCACAA', 'TGCGTACC'],
             'SI-3A-E8': ['AGCCTATG', 'CTAACGCA', 'GCTTACAT', 'TAGGGTGC'],
             'SI-3A-E9': ['AGTAAGCA', 'CCGGTAAT', 'GTATCTTC', 'TACCGCGG'],
             'SI-3A-F1': ['ATCGCTCC', 'CCGTACAG', 'GATAGGTA', 'TGACTAGT'],
             'SI-3A-F10': ['ATTCGTGC', 'CGCGTGCA', 'GAATACTG', 'TCGACAAT'],
             'SI-3A-F11': ['AGCAGTTA', 'CTTGTACC', 'GAACCCGG', 'TCGTAGAT'],
             'SI-3A-F12': ['AATTGAAC', 'CCAGTGGA', 'GTCCATTG', 'TGGACCCT'],
             'SI-3A-F2': ['ATGGTTAG', 'CATTGATA', 'GCAAACGC', 'TGCCCGCT'],
             'SI-3A-F3': ['AGTCTGTA', 'CAGAATAG', 'GCCTCCGT', 'TTAGGACC'],
             'SI-3A-F4': ['AACGACAC', 'CGTCCTCT', 'GCATGATA', 'TTGATGGG'],
             'SI-3A-F5': ['AACACAGC', 'CGGTTTAG', 'GTACGGCT', 'TCTGACTA'],
             'SI-3A-F6': ['ATGCCGGC', 'CCTAATTA', 'GACTTCCT', 'TGAGGAAG'],
             'SI-3A-F7': ['ACCCGAGA', 'CAAACTTT', 'GGTTAGAC', 'TTGGTCCG'],
             'SI-3A-F8': ['AGTTGGGA', 'CCAGAAAG', 'GTGCCCTC', 'TACATTCT'],
             'SI-3A-F9': ['AGTTAGTT', 'CACGCACG', 'GTACTTAA', 'TCGAGCGC'],
             'SI-3A-G1': ['ATGCGATT', 'CATATGCG', 'GGATACGA', 'TCCGCTAC'],
             'SI-3A-G10': ['ATACTGAG', 'CGGAGACT', 'GATGCCTC', 'TCCTATGA'],
             'SI-3A-G11': ['AGGGCGTT', 'CTATACGC', 'GCTCGTCA', 'TACATAAG'],
             'SI-3A-G12': ['ACCCGCAC', 'CATGCGTA', 'GTGATAGT', 'TGATATCG'],
             'SI-3A-G2': ['ATAACCTA', 'CGGTGAGC', 'GATCTTAT', 'TCCGAGCG'],
             'SI-3A-G3': ['ATGTCCAG', 'CGACGTCA', 'GCTATAGC', 'TACGAGTT'],
             'SI-3A-G4': ['AGCTTCTC', 'CCTGCGGT', 'GTACAACG', 'TAGAGTAA'],
             'SI-3A-G5': ['ATGAAGTA', 'CGCCGAAC', 'GAAGCTCG', 'TCTTTCGT'],
             'SI-3A-G6': ['AGCACTGG', 'CATTACAC', 'GTGCGACA', 'TCAGTGTT'],
             'SI-3A-G7': ['ATTACCGG', 'CAGTAATT', 'GCCGGTAA', 'TGACTGCC'],
             'SI-3A-G8': ['AAGTACTC', 'CTTGGAGA', 'GGAACTCT', 'TCCCTGAG'],
             'SI-3A-G9': ['AGTCTCAG', 'CAATGGCA', 'GCCGAAGT', 'TTGACTTC'],
             'SI-3A-H1': ['AAACTCAT', 'CGGGAGTA', 'GTCACAGG', 'TCTTGTCC'],
             'SI-3A-H10': ['ATTTCAGC', 'CGAGTGAT', 'GACCGCCA', 'TCGAATTG'],
             'SI-3A-H11': ['AGGATCGA', 'CACGATTC', 'GTATCGAG', 'TCTCGACT'],
             'SI-3A-H12': ['ACGAGTAG', 'CAATCCCT', 'GTCCAGGC', 'TGTGTATA'],
             'SI-3A-H2': ['AACGGTCA', 'CCGAACTC', 'GGTCCAAG', 'TTATTGGT'],
             'SI-3A-H3': ['ACACCTAA', 'CGTTTGGG', 'GACAAACC', 'TTGGGCTT'],
             'SI-3A-H4': ['ACTGGAGC', 'CGGTCGTG', 'GAAATCAA', 'TTCCATCT'],
             'SI-3A-H5': ['ATAGTATG', 'CCGCGTCT', 'GGCTCCAC', 'TATAAGGA'],
             'SI-3A-H6': ['AAGCATAA', 'CCCATCGC', 'GGTTGATG', 'TTAGCGCT'],
             'SI-3A-H7': ['AACGGGTG', 'CTAATTCT', 'GCTTCAAC', 'TGGCACGA'],
             'SI-3A-H8': ['AAGAGCGG', 'CTTGTTAT', 'GGCCCATC', 'TCATAGCA'],
             'SI-3A-H9': ['ACCTGCCA', 'CTTCATAC', 'GGAATATG', 'TAGGCGGT'],
             'SI-P03-A1': ['AAACGGCG', 'CCTACCAT', 'GGCGTTTC', 'TTGTAAGA'],
             'SI-P03-A10': ['ACAGCAAC', 'CGCAATTT', 'GAGTTGCG', 'TTTCGCGA'],
             'SI-P03-A11': ['ACCAGTCC', 'CTTTCCTT', 'GGACAGGG', 'TAGGTAAA'],
             'SI-P03-A12': ['ACTACTGT', 'CGGGAACG', 'GACCTCTC', 'TTATGGAA'],
             'SI-P03-A2': ['AGCCCTTT', 'CAAGTCCA', 'GTGAGAAG', 'TCTTAGGC'],
             'SI-P03-A3': ['AAAGCATA', 'CTGCAGCC', 'GCCTTTAT', 'TGTAGCGG'],
             'SI-P03-A4': ['AGAACGCC', 'CATGGCAG', 'GTCTTTGA', 'TCGCAATT'],
             'SI-P03-A5': ['ATTGGGAA', 'CAGTCTGG', 'GGCATACT', 'TCACACTC'],
             'SI-P03-A6': ['ACGGGACT', 'CTTTCGAC', 'GAACATGA', 'TGCATCTG'],
             'SI-P03-A7': ['AGGTCATA', 'CTCATCAT', 'GCTGAGGG', 'TAACGTCC'],
             'SI-P03-A8': ['ATGATACG', 'CCACAGAA', 'GACTGTTC', 'TGTGCCGT'],
             'SI-P03-A9': ['ACAACTTG', 'CTCCAACA', 'GAGTGCGT', 'TGTGTGAC'],
             'SI-P03-B1': ['AGGCTACC', 'CTAGCTGT', 'GCCAACAA', 'TATTGGTG'],
             'SI-P03-B10': ['ACCATTAA', 'CTGGACGT', 'GAACGGTC', 'TGTTCACG'],
             'SI-P03-B11': ['ATGGTCGC', 'CGACATAG', 'GATTCGCT', 'TCCAGATA'],
             'SI-P03-B12': ['ACGCTTGG', 'CGCTACAT', 'GAAAGACA', 'TTTGCGTC'],
             'SI-P03-B2': ['AAGTTGAT', 'CCCACCCA', 'GGTCGAGC', 'TTAGATTG'],
             'SI-P03-B3': ['ATTGGACG', 'CAGCTTAC', 'GGCAAGGA', 'TCATCCTT'],
             'SI-P03-B4': ['AGGGACTG', 'CCTCTAAC', 'GACAGGCT', 'TTATCTGA'],
             'SI-P03-B5': ['ATCGTACT', 'CATCAGTG', 'GGGACTAC', 'TCATGCGA'],
             'SI-P03-B6': ['AACGCGAA', 'CTATTTGG', 'GCGCACCT', 'TGTAGATC'],
             'SI-P03-B7': ['AGGGATGA', 'CTTCTGTT', 'GAATGCAC', 'TCCACACG'],
             'SI-P03-B8': ['ACGTTCAC', 'CAAGGTCT', 'GTTAAGTG', 'TGCCCAGA'],
             'SI-P03-B9': ['AAGCGTGT', 'CTTGACCG', 'GCCACGTA', 'TGATTAAC'],
             'SI-P03-C1': ['AGACTTTC', 'CCGAGGCA', 'GATGCAGT', 'TTCTACAG'],
             'SI-P03-C10': ['ATCTGATC', 'CGTGCTAA', 'GAGAAGGG', 'TCACTCCT'],
             'SI-P03-C11': ['ACCGAACA', 'CGACTCTT', 'GTTTGTGG', 'TAGACGAC'],
             'SI-P03-C12': ['ATCCGGCA', 'CCGTTATG', 'GGTAATGT', 'TAAGCCAC'],
             'SI-P03-C2': ['AATCACTA', 'CCGAGAAC', 'GTAGTGCG', 'TGCTCTGT'],
             'SI-P03-C3': ['ACGTTACA', 'CGTAGGTT', 'GACGACGG', 'TTACCTAC'],
             'SI-P03-C4': ['ACATTGGC', 'CTTAGTCA', 'GAGCCCAT', 'TGCGAATG'],
             'SI-P03-C5': ['ATGCATTC', 'CACTGACT', 'GGTACGGG', 'TCAGTCAA'],
             'SI-P03-C6': ['ACTCAGAC', 'CGCTCAGG', 'GAGGTTTA', 'TTAAGCCT'],
             'SI-P03-C7': ['ACACCGGG', 'CATAATCC', 'GGCGGAAT', 'TTGTTCTA'],
             'SI-P03-C8': ['AGCTCGAG', 'CAGGAAGA', 'GCACGTTT', 'TTTATCCC'],
             'SI-P03-C9': ['AGATCGGT', 'CATCGTCG', 'GTCATATA', 'TCGGACAC'],
             'SI-P03-D1': ['AGCTGCGT', 'CAACCATC', 'GTGGAGCA', 'TCTATTAG'],
             'SI-P03-D10': ['AGATAACA', 'CTTATTTG', 'GCGGGCAT', 'TACCCGGC'],
             'SI-P03-D11': ['ATATGAGA', 'CACCTCAG', 'GCTACTTC', 'TGGGAGCT'],
             'SI-P03-D12': ['AGAAACGT', 'CACTCAAC', 'GCTGTGTA', 'TTGCGTCG'],
             'SI-P03-D2': ['ACATTCCG', 'CTGCGGTA', 'GACACAAT', 'TGTGATGC'],
             'SI-P03-D3': ['ACTTCACT', 'CGAAGTTG', 'GAGCACGC', 'TTCGTGAA'],
             'SI-P03-D4': ['AAATCGTC', 'CTTCGAAT', 'GCGATCGG', 'TGCGATCA'],
             'SI-P03-D5': ['AGACGGAT', 'CCTTTAGA', 'GTCGACTC', 'TAGACTCG'],
             'SI-P03-D6': ['ATGCCAAA', 'CCTTATCG', 'GAAGTCTT', 'TGCAGGGC'],
             'SI-P03-D7': ['AACTTAGA', 'CCGGATCC', 'GGTCGCAT', 'TTAACGTG'],
             'SI-P03-D8': ['AATCTTTG', 'CTCAAGAC', 'GGATGAGT', 'TCGGCCCA'],
             'SI-P03-D9': ['ACCTACTG', 'CAAGGGAC', 'GGGACACA', 'TTTCTTGT'],
             'SI-P03-E1': ['ACGAAAGC', 'CGCCCGTA', 'GTTTGCCT', 'TAAGTTAG'],
             'SI-P03-E10': ['ATTGTTTC', 'CGCAGGAG', 'GCACCAGT', 'TAGTACCA'],
             'SI-P03-E11': ['ATCGCCAT', 'CATAAAGG', 'GGGTTTCC', 'TCACGGTA'],
             'SI-P03-E12': ['ACGCGGAA', 'CGCTATCC', 'GTTGCATG', 'TAAATCGT'],
             'SI-P03-E2': ['AGGCTGGT', 'CACAACTA', 'GTTGGTCC', 'TCATCAAG'],
             'SI-P03-E3': ['AACAAGTC', 'CGGCTCCA', 'GTATGTAT', 'TCTGCAGG'],
             'SI-P03-E4': ['AGCTGACG', 'CCGGTGTC', 'GTAAACAT', 'TATCCTGA'],
             'SI-P03-E5': ['ATCCAAGG', 'CCGTTGAA', 'GGAAGCTC', 'TATGCTCT'],
             'SI-P03-E6': ['ATTGAAAC', 'CAGCCCGA', 'GCCATTTG', 'TGATGGCT'],
             'SI-P03-E7': ['AAGACGTG', 'CCATGTGT', 'GTTCACAA', 'TGCGTACC'],
             'SI-P03-E8': ['AGCCTATG', 'CTAACGCA', 'GCTTACAT', 'TAGGGTGC'],
             'SI-P03-E9': ['AGTAAGCA', 'CCGGTAAT', 'GTATCTTC', 'TACCGCGG'],
             'SI-P03-F1': ['ATCGCTCC', 'CCGTACAG', 'GATAGGTA', 'TGACTAGT'],
             'SI-P03-F10': ['ATTCGTGC', 'CGCGTGCA', 'GAATACTG', 'TCGACAAT'],
             'SI-P03-F11': ['AGCAGTTA', 'CTTGTACC', 'GAACCCGG', 'TCGTAGAT'],
             'SI-P03-F12': ['AATTGAAC', 'CCAGTGGA', 'GTCCATTG', 'TGGACCCT'],
             'SI-P03-F2': ['ATGGTTAG', 'CATTGATA', 'GCAAACGC', 'TGCCCGCT'],
             'SI-P03-F3': ['AGTCTGTA', 'CAGAATAG', 'GCCTCCGT', 'TTAGGACC'],
             'SI-P03-F4': ['AACGACAC', 'CGTCCTCT', 'GCATGATA', 'TTGATGGG'],
             'SI-P03-F5': ['AACACAGC', 'CGGTTTAG', 'GTACGGCT', 'TCTGACTA'],
             'SI-P03-F6': ['ATGCCGGC', 'CCTAATTA', 'GACTTCCT', 'TGAGGAAG'],
             'SI-P03-F7': ['ACCCGAGA', 'CAAACTTT', 'GGTTAGAC', 'TTGGTCCG'],
             'SI-P03-F8': ['AGTTGGGA', 'CCAGAAAG', 'GTGCCCTC', 'TACATTCT'],
             'SI-P03-F9': ['AGTTAGTT', 'CACGCACG', 'GTACTTAA', 'TCGAGCGC'],
             'SI-P03-G1': ['ATGCGATT', 'CATATGCG', 'GGATACGA', 'TCCGCTAC'],
             'SI-P03-G10': ['ATACTGAG', 'CGGAGACT', 'GATGCCTC', 'TCCTATGA'],
             'SI-P03-G11': ['AGGGCGTT', 'CTATACGC', 'GCTCGTCA', 'TACATAAG'],
             'SI-P03-G12': ['ACCCGCAC', 'CATGCGTA', 'GTGATAGT', 'TGATATCG'],
             'SI-P03-G2': ['ATAACCTA', 'CGGTGAGC', 'GATCTTAT', 'TCCGAGCG'],
             'SI-P03-G3': ['ATGTCCAG', 'CGACGTCA', 'GCTATAGC', 'TACGAGTT'],
             'SI-P03-G4': ['AGCTTCTC', 'CCTGCGGT', 'GTACAACG', 'TAGAGTAA'],
             'SI-P03-G5': ['ATGAAGTA', 'CGCCGAAC', 'GAAGCTCG', 'TCTTTCGT'],
             'SI-P03-G6': ['AGCACTGG', 'CATTACAC', 'GTGCGACA', 'TCAGTGTT'],
             'SI-P03-G7': ['ATTACCGG', 'CAGTAATT', 'GCCGGTAA', 'TGACTGCC'],
             'SI-P03-G8': ['AAGTACTC', 'CTTGGAGA', 'GGAACTCT', 'TCCCTGAG'],
             'SI-P03-G9': ['AGTCTCAG', 'CAATGGCA', 'GCCGAAGT', 'TTGACTTC'],
             'SI-P03-H1': ['AAACTCAT', 'CGGGAGTA', 'GTCACAGG', 'TCTTGTCC'],
             'SI-P03-H10': ['ATTTCAGC', 'CGAGTGAT', 'GACCGCCA', 'TCGAATTG'],
             'SI-P03-H11': ['AGGATCGA', 'CACGATTC', 'GTATCGAG', 'TCTCGACT'],
             'SI-P03-H12': ['ACGAGTAG', 'CAATCCCT', 'GTCCAGGC', 'TGTGTATA'],
             'SI-P03-H2': ['AACGGTCA', 'CCGAACTC', 'GGTCCAAG', 'TTATTGGT'],
             'SI-P03-H3': ['ACACCTAA', 'CGTTTGGG', 'GACAAACC', 'TTGGGCTT'],
             'SI-P03-H4': ['ACTGGAGC', 'CGGTCGTG', 'GAAATCAA', 'TTCCATCT'],
             'SI-P03-H5': ['ATAGTATG', 'CCGCGTCT', 'GGCTCCAC', 'TATAAGGA'],
             'SI-P03-H6': ['AAGCATAA', 'CCCATCGC', 'GGTTGATG', 'TTAGCGCT'],
             'SI-P03-H7': ['AACGGGTG', 'CTAATTCT', 'GCTTCAAC', 'TGGCACGA'],
             'SI-P03-H8': ['AAGAGCGG', 'CTTGTTAT', 'GGCCCATC', 'TCATAGCA'],
             'SI-P03-H9': ['ACCTGCCA', 'CTTCATAC', 'GGAATATG', 'TAGGCGGT']}



import h5py
import numpy as np
import pickle
import zlib
import os
from typing import *


def _obj2uint(obj: object, compression: int=9, protocol: int=2) -> np.ndarray:
    """Transform a python object in a numpy array of uint8
    
    Arguments
    ---------
    obj: object
        The object to encode
    compression: int, default=9
        the compression level of ``zlib``
    protocol: int, default=2
        the protocol used by ``pickle.dumps``
    
    Returns
    -------
    An array encoding in bytes (uint8) the object pickled
    """
    zstr = zlib.compress(pickle.dumps(obj, protocol=protocol), compression)
    return np.fromstring(zstr, dtype=np.uint8)


def _uint2obj(uint: np.ndarray) -> object:
    """Recover the original object encoded in a uint8 numpy array
    
    Arguments
    ---------
    uint: np.ndarray
        A uint8 numpy array generated by the ``_obj2uint`` function
    
    Returns
    -------
    A python object that was encoded in the numpy array
    """
    return pickle.loads(zlib.decompress(uint.tobytes()))


def dump_hdf5(obj: object, filename: str,
              data_compression: int=7, chunks: Tuple =(2048, 2048),
              noarray_compression: int=9, pickle_protocol: int=2) -> None:
    """Dump all attribute of a python object to hdf5
    
    Arguments
    ---------
    obj: object
        a python object
    filename: str
        the name of the file to be saved
    data_compression: int
        the level of compression used by hdf5
    chunks: Tuple, default=(2048, 2048)
        The size of the chunks to be used for compression/random access
    noarray_compression: int, default=9
        the compression level of ``zlib``, used when the attribute is not an array
    pickle_protocol: int, default=2
        the protocol used by ``pickle.dumps`` when the attribute is not an array
    
    Returns
    -------
    Nothing but it creates a file ``filename``
    """
    if os.path.isfile(filename):
        os.remove(filename)
    _file = h5py.File(filename, "w")
    for k in obj.__dict__.keys():
        attribute = getattr(obj, k)
        if type(attribute) is not np.ndarray:
            serialized = _obj2uint(attribute, compression=noarray_compression, protocol=pickle_protocol)
            _file.create_dataset(f"&{k}", data=serialized,
                                 chunks=tuple((min(1024, len(serialized)),)),
                                 compression="gzip", compression_opts=data_compression,
                                 fletcher32=False, shuffle=False)
        else:
            chunk_size = tuple((min(chunks[i], attribute.shape[i]) for i in range(len(attribute.shape))))
            _file.create_dataset(k, data=attribute, chunks=chunk_size,
                                 compression="gzip", compression_opts=data_compression,
                                 fletcher32=False, shuffle=False)
    _file.close()


def load_hdf5(filename: str, obj_class: Type[object]) -> object:
    """Dump all attribute of a python object to hdf5
    
    Arguments
    ---------
    filename:
        the name of the file to be loaded
    obj_class:
        the type of object to be generated
    
    Returns
    -------
    An object that has been filled with the attributes stored in the hdf5 file
    
    Note
    ----
    The trick to create an empty object was adapted from Guido van Rossum answer:
    https://stackoverflow.com/questions/2168964/python-creating-class-instance-without-calling-initializer
    """
    
    obj = obj_class.__new__(obj_class)
    _file = h5py.File(filename, "r")
    for k in _file.keys():
        if k.startswith("&"):
            setattr(obj, k[1:], _uint2obj(_file[k][:]))
        else:
            setattr(obj, k, _file[k][:])
    _file.close()
    return obj



from typing import *
import velocyto as vcy
import numpy as np
import abc


class Logic(metaclass=abc.ABCMeta):
    """Base class from wich all the logics should inherit
    """

    def __init__(self) -> None:
        self.name = "Logic"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return []

    @property
    def stranded(self) -> bool:
        return True

    @property
    def perform_validation_markup(self) -> bool:
        return True

    @property
    def accept_discordant(self) -> bool:
        return False

    @abc.abstractmethod  # This needs to be overridden
    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> Union[None, int]:
        """This methods will have to countain the core operations of the logic to attribute a molecule to one of the cathergories
        
        Arguments
        ---------
        molitem: vcy.Molitem
            The :py:class:`vcy.Molitem` object to be considered by the logic
        cell_bcidx: int
            The cell index in the memory buffers below
        dict_layers_columns: Dict[str, np.ndarray]
            A dictionary mapping the name of a layer with the memory buffer that will be saved in the loom file after counting
        geneid2ix: Dict[str, int]
            Dictionary containing the Acession of the genes mapping to its column index position
        
        Returns
        -------
        Nothing but it adds the molecule to the appropriate layer (or does not count at all)

        """
        # NOTE I need to generalize this to any set of layers
        return None


class Permissive10X(Logic):
    """Permissive logic for 10X Genomics chemistry

    This logic differs from the other 10x Logics because:
    - singletons if the fall in not validated introns are COUNTED UNSPLICED
    - singletons if the fall in validated introns are COUNTED UNSPLICED
    - non-singletons if are supported by not validated introns are COUNTED UNSPLICED
    - non-singletons if are supported by validated introns are COUNTED UNSPLICED
    """

    def __init__(self) -> None:
        self.name = "Permissive10X"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return ["spliced", "unspliced", "ambiguous"]

    @property
    def stranded(self) -> bool:
        return True

    @property
    def perform_validation_markup(self) -> bool:
        return True

    @property
    def accept_discordant(self) -> bool:
        return False

    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> int:
        # NOTE This can be simplified qyuite a bit, without loss of acuracy!
        # The hits are not compatible with any annotated transcript model
        spliced = dict_layers_columns["spliced"]
        unspliced = dict_layers_columns["unspliced"]
        ambiguous = dict_layers_columns["ambiguous"]
        # The hits are not compatible with any annotated transcript model
        if len(molitem.mappings_record) == 0:
            return 2
        # Compatible with one or more transcript models:
        else:
            # Check that there are not different possible genes ??
            if len(set(i.geneid for i in molitem.mappings_record.keys())) == 1:
                gene_check: Set[str] = set()
                
                has_onlyintron_model = 0
                has_only_span_exin_model = 1
                has_onlyintron_and_valid_model = 0
                has_valid_mixed_model = 0
                has_invalid_mixed_model = 0
                has_onlyexo_model = 0
                has_mixed_model = 0
                multi_gene = 0
                for transcript_model, segments_list in molitem.mappings_record.items():
                    gene_check.add(transcript_model.geneid)
                    if len(gene_check) > 1:
                        multi_gene = 1
                    has_introns = 0
                    has_exons = 0
                    has_exseg_with_spliced_flag = 0
                    has_validated_intron = 0
                    has_exin_intron_span = 0
                    has_non3prime = 0
                    for segment_match in segments_list:
                        if segment_match.maps_to_intron:
                            has_introns = 1
                            if segment_match.feature.is_validated:
                                has_validated_intron = 1
                                if segment_match.feature.end_overlaps_with_part_of(segment_match.segment):
                                    downstream_exon = segment_match.feature.get_downstream_exon()
                                    if downstream_exon.start_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                                if segment_match.feature.start_overlaps_with_part_of(segment_match.segment):
                                    upstream_exon = segment_match.feature.get_upstream_exon()
                                    if upstream_exon.end_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                        elif segment_match.maps_to_exon:
                            has_exons = 1
                            if not segment_match.feature.is_last_3prime:
                                has_non3prime = 1
                            if segment_match.is_spliced:
                                has_exseg_with_spliced_flag = 1
                    if has_validated_intron and not has_exons:
                        has_onlyintron_and_valid_model = 1
                    if has_introns and not has_exons:
                        has_onlyintron_model = 1
                    if has_exons and not has_introns:
                        has_onlyexo_model = 1
                    if has_exons and has_introns and not has_validated_intron and not has_exin_intron_span:
                        has_invalid_mixed_model = 1
                        has_mixed_model = 1
                    if has_exons and has_introns and has_validated_intron and not has_exin_intron_span:
                        has_valid_mixed_model = 1
                        has_mixed_model = 1
                    if not has_exin_intron_span:
                        has_only_span_exin_model = 0
                        
                if multi_gene:
                    # Many genes are compatible with the observation, do not count
                    return 1
                else:
                    if not len(molitem.mappings_record):
                        # No gene is compatible with the observation, do not count
                        return 2
                    else:
                        if has_onlyexo_model and not has_onlyintron_model and not has_mixed_model:
                            # More common situation, normal exonic read, count as spliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spliced[gene_ix, cell_bcidx] += 1
                            return 0
                        if has_only_span_exin_model:
                            # All the compatible transcript models have spanning exon-intron boundaries, count unspliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return 0
                        if has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return 0
                            else:
                                # Non-singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return 0
                        if has_onlyintron_model and not has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in non-validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return 0
                            else:
                                # Non-singleton in non-validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return 0
                        if has_invalid_mixed_model and not has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Not validated and mapping to exon and introns, happens rarely in 10X / irrelevant. Count anyways
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return 0
                        if has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Validated and mapping to exon and introns, happens rarely in 10X. Count as unspliced.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return 0
                        if has_onlyintron_model and has_onlyexo_model and not has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping, most common case! Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return 0
                        if has_onlyintron_model and not has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return 0
                        if not has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return 0
                        if has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
        return 3
                    

class Intermediate10X(Logic):
    """ValidatedIntrons logic for 10X Genomics chemistry
    
    This differs from the other 10x Logics because:
    - singletons if the fall in not validated introns are DISCARDED
    - singletons if the fall in validated introns are COUNTED UNSPLICED
    - non-singletons if are supported by not validated introns are COUNTED UNSPLICED
    - non-singletons if are supported by validated introns are COUNTED UNSPLICED

    """

    def __init__(self) -> None:
        self.name = "Intermediate10X"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return ["spliced", "unspliced", "ambiguous"]

    @property
    def stranded(self) -> bool:
        return True

    @property
    def perform_validation_markup(self) -> bool:
        return True

    @property
    def accept_discordant(self) -> bool:
        return False

    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> None:
        # NOTE This can be simplified qyuite a bit, without loss of acuracy!
        # The hits are not compatible with any annotated transcript model
        spliced = dict_layers_columns["spliced"]
        unspliced = dict_layers_columns["unspliced"]
        ambiguous = dict_layers_columns["ambiguous"]
        # The hits are not compatible with any annotated transcript model
        if len(molitem.mappings_record) == 0:
            return
        # Compatible with one or more transcript models:
        else:
            # Check that there are not different possible genes ??
            if len(set(i.geneid for i in molitem.mappings_record.keys())) == 1:
                gene_check: Set[str] = set()
                
                has_onlyintron_model = 0
                has_only_span_exin_model = 1
                has_onlyintron_and_valid_model = 0
                has_valid_mixed_model = 0
                has_invalid_mixed_model = 0
                has_onlyexo_model = 0
                has_mixed_model = 0
                multi_gene = 0
                for transcript_model, segments_list in molitem.mappings_record.items():
                    gene_check.add(transcript_model.geneid)
                    if len(gene_check) > 1:
                        multi_gene = 1
                    has_introns = 0
                    has_exons = 0
                    has_exseg_with_spliced_flag = 0
                    has_validated_intron = 0
                    has_exin_intron_span = 0
                    has_non3prime = 0
                    for segment_match in segments_list:
                        if segment_match.maps_to_intron:
                            has_introns = 1
                            if segment_match.feature.is_validated:
                                has_validated_intron = 1
                                if segment_match.feature.end_overlaps_with_part_of(segment_match.segment):
                                    downstream_exon = segment_match.feature.get_downstream_exon()
                                    if downstream_exon.start_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                                if segment_match.feature.start_overlaps_with_part_of(segment_match.segment):
                                    upstream_exon = segment_match.feature.get_upstream_exon()
                                    if upstream_exon.end_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                        elif segment_match.maps_to_exon:
                            has_exons = 1
                            if not segment_match.feature.is_last_3prime:
                                has_non3prime = 1
                            if segment_match.is_spliced:
                                has_exseg_with_spliced_flag = 1
                    if has_validated_intron and not has_exons:
                        has_onlyintron_and_valid_model = 1
                    if has_introns and not has_exons:
                        has_onlyintron_model = 1
                    if has_exons and not has_introns:
                        has_onlyexo_model = 1
                    if has_exons and has_introns and not has_validated_intron and not has_exin_intron_span:
                        has_invalid_mixed_model = 1
                        has_mixed_model = 1
                    if has_exons and has_introns and has_validated_intron and not has_exin_intron_span:
                        has_valid_mixed_model = 1
                        has_mixed_model = 1
                    if not has_exin_intron_span:
                        has_only_span_exin_model = 0
                        
                if multi_gene:
                    # many genes are compatible with the observation, do not count
                    return
                else:
                    if not len(molitem.mappings_record):
                        # no gene is compatible with the observation, do not count
                        return
                    else:
                        if has_onlyexo_model and not has_onlyintron_model and not has_mixed_model:
                            # More common situation, normal exonic read, count as spliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_only_span_exin_model:
                            # All the compatible transcript models have spanning exon-intron boundaries, count unspliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                            else:
                                # Non singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                        if has_onlyintron_model and not has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in non-validated intron
                                return
                            else:
                                # Non-singleton in non-validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                        if has_invalid_mixed_model and not has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Not validated and mapping to exon and introns, happens rarely in 10X / irrelevant.
                            return
                        if has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Validated and mapping to exon and introns, happens rarely in 10X. Count as unspliced.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and not has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping, most common case! Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and not has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if not has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return


class ValidatedIntrons10X(Logic):
    """ValidatedIntrons logic for 10X Genomics chemistry
    
    This differs from the other 10x Logics because:
    - singletons if the fall in not validated introns are DISCARDED
    - singletons if the fall in validated introns are COUNTED UNSPLICED
    - non-singletons if are supported by not validated introns are DISCARDED
    - non-singletons if are supported by validated introns are COUNTED UNSPLICED

    """

    def __init__(self) -> None:
        self.name = "ValidatedIntrons10X"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return ["spliced", "unspliced", "ambiguous"]

    @property
    def stranded(self) -> bool:
        return True

    @property
    def perform_validation_markup(self) -> bool:
        return True

    @property
    def accept_discordant(self) -> bool:
        return False

    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> None:
        # NOTE This can be simplified qyuite a bit, without loss of acuracy!
        # The hits are not compatible with any annotated transcript model
        spliced = dict_layers_columns["spliced"]
        unspliced = dict_layers_columns["unspliced"]
        ambiguous = dict_layers_columns["ambiguous"]
        # The hits are not compatible with any annotated transcript model
        if len(molitem.mappings_record) == 0:
            return
        # Compatible with one or more transcript models:
        else:
            # Check that there are not different possible genes ??
            if len(set(i.geneid for i in molitem.mappings_record.keys())) == 1:
                gene_check: Set[str] = set()
                
                has_onlyintron_model = 0
                has_only_span_exin_model = 1
                has_onlyintron_and_valid_model = 0
                has_valid_mixed_model = 0
                has_invalid_mixed_model = 0
                has_onlyexo_model = 0
                has_mixed_model = 0
                multi_gene = 0
                for transcript_model, segments_list in molitem.mappings_record.items():
                    gene_check.add(transcript_model.geneid)
                    if len(gene_check) > 1:
                        multi_gene = 1
                    has_introns = 0
                    has_exons = 0
                    has_exseg_with_spliced_flag = 0
                    has_validated_intron = 0
                    has_exin_intron_span = 0
                    has_non3prime = 0
                    for segment_match in segments_list:
                        if segment_match.maps_to_intron:
                            has_introns = 1
                            if segment_match.feature.is_validated:
                                has_validated_intron = 1
                                if segment_match.feature.end_overlaps_with_part_of(segment_match.segment):
                                    downstream_exon = segment_match.feature.get_downstream_exon()
                                    if downstream_exon.start_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                                if segment_match.feature.start_overlaps_with_part_of(segment_match.segment):
                                    upstream_exon = segment_match.feature.get_upstream_exon()
                                    if upstream_exon.end_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                        elif segment_match.maps_to_exon:
                            has_exons = 1
                            if not segment_match.feature.is_last_3prime:
                                has_non3prime = 1
                            if segment_match.is_spliced:
                                has_exseg_with_spliced_flag = 1
                    if has_validated_intron and not has_exons:
                        has_onlyintron_and_valid_model = 1
                    if has_introns and not has_exons:
                        has_onlyintron_model = 1
                    if has_exons and not has_introns:
                        has_onlyexo_model = 1
                    if has_exons and has_introns and not has_validated_intron and not has_exin_intron_span:
                        has_invalid_mixed_model = 1
                        has_mixed_model = 1
                    if has_exons and has_introns and has_validated_intron and not has_exin_intron_span:
                        has_valid_mixed_model = 1
                        has_mixed_model = 1
                    if not has_exin_intron_span:
                        has_only_span_exin_model = 0
                        
                if multi_gene:
                    # many genes are compatible with the observation, do not count
                    return
                else:
                    if not len(molitem.mappings_record):
                        # no gene is compatible with the observation, do not count
                        return
                    else:
                        if has_onlyexo_model and not has_onlyintron_model and not has_mixed_model:
                            # More common situation, normal exonic read, count as spliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_only_span_exin_model:
                            # All the compatible transcript models have spanning exon-intron boundaries, count unspliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                            else:
                                # Non singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                        if has_onlyintron_model and not has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in non-validated intron
                                return
                            else:
                                # Non-singleton in non-validated intron
                                return
                        if has_invalid_mixed_model and not has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Not validated and mapping to exon and introns, happens rarely in 10X / irrelevant.
                            return
                        if has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Validated and mapping to exon and introns, happens rarely in 10X. Count as unspliced.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and not has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping, most common case! Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and not has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if not has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return


class Stricter10X(Logic):
    """Stricter logic for 10X Genomics chemistry

    This differ from the other 10x Logics because:
    - singletons if the fall in not validated introns are DISCARDED
    - singletons if the fall in validated introns are DISCARDED
    - non-singletons if are supported by not validated introns are DISCARDED
    - non-singletons if are supported by validated introns are COUNTED UNSPLICED

    """

    def __init__(self) -> None:
        self.name = "Stricter10X"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return ["spliced", "unspliced", "ambiguous"]

    @property
    def stranded(self) -> bool:
        return True

    @property
    def perform_validation_markup(self) -> bool:
        return True

    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> None:
        # NOTE This can be simplified qyuite a bit, without loss of acuracy!
        # The hits are not compatible with any annotated transcript model
        spliced = dict_layers_columns["spliced"]
        unspliced = dict_layers_columns["unspliced"]
        ambiguous = dict_layers_columns["ambiguous"]
        # The hits are not compatible with any annotated transcript model
        if len(molitem.mappings_record) == 0:
            return
        # Compatible with one or more transcript models:
        else:
            # Check that there are not different possible genes ??
            if len(set(i.geneid for i in molitem.mappings_record.keys())) == 1:
                gene_check: Set[str] = set()
                
                has_onlyintron_model = 0
                has_only_span_exin_model = 1
                has_onlyintron_and_valid_model = 0
                has_valid_mixed_model = 0
                has_invalid_mixed_model = 0
                has_onlyexo_model = 0
                has_mixed_model = 0
                multi_gene = 0
                for transcript_model, segments_list in molitem.mappings_record.items():
                    gene_check.add(transcript_model.geneid)
                    if len(gene_check) > 1:
                        multi_gene = 1
                    has_introns = 0
                    has_exons = 0
                    has_exseg_with_spliced_flag = 0
                    has_validated_intron = 0
                    has_exin_intron_span = 0
                    has_non3prime = 0
                    for segment_match in segments_list:
                        if segment_match.maps_to_intron:
                            has_introns = 1
                            if segment_match.feature.is_validated:
                                has_validated_intron = 1
                                if segment_match.feature.end_overlaps_with_part_of(segment_match.segment):
                                    downstream_exon = segment_match.feature.get_downstream_exon()
                                    if downstream_exon.start_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                                if segment_match.feature.start_overlaps_with_part_of(segment_match.segment):
                                    upstream_exon = segment_match.feature.get_upstream_exon()
                                    if upstream_exon.end_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                        elif segment_match.maps_to_exon:
                            has_exons = 1
                            if not segment_match.feature.is_last_3prime:
                                has_non3prime = 1
                            if segment_match.is_spliced:
                                has_exseg_with_spliced_flag = 1
                    if has_validated_intron and not has_exons:
                        has_onlyintron_and_valid_model = 1
                    if has_introns and not has_exons:
                        has_onlyintron_model = 1
                    if has_exons and not has_introns:
                        has_onlyexo_model = 1
                    if has_exons and has_introns and not has_validated_intron and not has_exin_intron_span:
                        has_invalid_mixed_model = 1
                        has_mixed_model = 1
                    if has_exons and has_introns and has_validated_intron and not has_exin_intron_span:
                        has_valid_mixed_model = 1
                        has_mixed_model = 1
                    if not has_exin_intron_span:
                        has_only_span_exin_model = 0
                        
                if multi_gene:
                    # many genes are compatible with the observation, do not count
                    return
                else:
                    if not len(molitem.mappings_record):
                        # no gene is compatible with the observation, do not count
                        return
                    else:
                        if has_onlyexo_model and not has_onlyintron_model and not has_mixed_model:
                            # More common situation, normal exonic read, count as spliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_only_span_exin_model:
                            # All the compatible transcript models have spanning exon-intron boundaries, count unspliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in validated intron, do not count
                                return
                            else:
                                # Non singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                        if has_onlyintron_model and not has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in non-validated intron
                                return
                            else:
                                # Non-singleton in non-validated intron
                                return
                        if has_invalid_mixed_model and not has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Not validated and mapping to exon and introns, happens rarely in 10X / irrelevant.
                            return
                        if has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Validated and mapping to exon and introns, happens rarely in 10X. Count as unspliced.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and not has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping, most common case! Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and not has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if not has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return


class ObservedSpanning10X(Logic):
    """ObservedSpanning10X logic for 10X Genomics chemistry

    This differ from the other 10x Logics because:
    - singletons if the fall in not validated introns are DISCARDED
    - singletons if the fall in validated introns are DISCARDED
    - non-singletons if are supported by not validated introns are DISCARDED
    - non-singletons if are supported by validated introns are DISCARDED
    - Therefore only the observed intron spanning reads are counted as UNSPLICED
    """

    def __init__(self) -> None:
        self.name = "ObservedSpanning10X"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return ["spliced", "unspliced", "ambiguous"]

    @property
    def stranded(self) -> bool:
        return True

    @property
    def perform_validation_markup(self) -> bool:
        return True

    @property
    def accept_discordant(self) -> bool:
        return False

    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> None:
        # NOTE This can be simplified qyuite a bit, without loss of acuracy!
        # The hits are not compatible with any annotated transcript model
        spliced = dict_layers_columns["spliced"]
        unspliced = dict_layers_columns["unspliced"]
        ambiguous = dict_layers_columns["ambiguous"]
        # The hits are not compatible with any annotated transcript model
        if len(molitem.mappings_record) == 0:
            return
        # Compatible with one or more transcript models:
        else:
            # Check that there are not different possible genes ??
            if len(set(i.geneid for i in molitem.mappings_record.keys())) == 1:
                gene_check: Set[str] = set()
                
                has_onlyintron_model = 0
                has_only_span_exin_model = 1
                has_onlyintron_and_valid_model = 0
                has_valid_mixed_model = 0
                has_invalid_mixed_model = 0
                has_onlyexo_model = 0
                has_mixed_model = 0
                multi_gene = 0
                for transcript_model, segments_list in molitem.mappings_record.items():
                    gene_check.add(transcript_model.geneid)
                    if len(gene_check) > 1:
                        multi_gene = 1
                    has_introns = 0
                    has_exons = 0
                    has_exseg_with_spliced_flag = 0
                    has_validated_intron = 0
                    has_exin_intron_span = 0
                    has_non3prime = 0
                    for segment_match in segments_list:
                        if segment_match.maps_to_intron:
                            has_introns = 1
                            if segment_match.feature.is_validated:
                                has_validated_intron = 1
                                if segment_match.feature.end_overlaps_with_part_of(segment_match.segment):
                                    downstream_exon = segment_match.feature.get_downstream_exon()
                                    if downstream_exon.start_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                                if segment_match.feature.start_overlaps_with_part_of(segment_match.segment):
                                    upstream_exon = segment_match.feature.get_upstream_exon()
                                    if upstream_exon.end_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                        elif segment_match.maps_to_exon:
                            has_exons = 1
                            if not segment_match.feature.is_last_3prime:
                                has_non3prime = 1
                            if segment_match.is_spliced:
                                has_exseg_with_spliced_flag = 1
                    if has_validated_intron and not has_exons:
                        has_onlyintron_and_valid_model = 1
                    if has_introns and not has_exons:
                        has_onlyintron_model = 1
                    if has_exons and not has_introns:
                        has_onlyexo_model = 1
                    if has_exons and has_introns and not has_validated_intron and not has_exin_intron_span:
                        has_invalid_mixed_model = 1
                        has_mixed_model = 1
                    if has_exons and has_introns and has_validated_intron and not has_exin_intron_span:
                        has_valid_mixed_model = 1
                        has_mixed_model = 1
                    if not has_exin_intron_span:
                        has_only_span_exin_model = 0
                        
                if multi_gene:
                    # Many genes are compatible with the observation, do not count
                    return
                else:
                    if not len(molitem.mappings_record):
                        # No gene is compatible with the observation, do not count
                        return
                    else:
                        if has_onlyexo_model and not has_onlyintron_model and not has_mixed_model:
                            # More common situation, normal exonic read, count as spliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_only_span_exin_model:
                            # All the compatible transcript models have spanning exon-intron boundaries, count unspliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in validated intron, do not count
                                return
                            else:
                                # Non-singleton in validated intron, do not count
                                return
                        if has_onlyintron_model and not has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in non-validated intron
                                return
                            else:
                                # Non-singleton in non-validated intron
                                return
                        if has_invalid_mixed_model and not has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Not validated and mapping to exon and introns, happens rarely in 10X / irrelevant.
                            return
                        if has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Validated and mapping to exon and introns, happens rarely in 10X. Count as unspliced.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and not has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping, most common case! Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and not has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if not has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return


class Discordant10X(Logic):
    """Just a test
    """

    def __init__(self) -> None:
        self.name = "Discordant10X"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return ["spliced", "unspliced", "ambiguous"]

    @property
    def stranded(self) -> bool:
        return True

    @property
    def perform_validation_markup(self) -> bool:
        return True

    @property
    def accept_discordant(self) -> bool:
        return True

    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> None:
        # NOTE This can be simplified qyuite a bit, without loss of acuracy!
        # The hits are not compatible with any annotated transcript model
        spliced = dict_layers_columns["spliced"]
        unspliced = dict_layers_columns["unspliced"]
        ambiguous = dict_layers_columns["ambiguous"]
        # The hits are not compatible with any annotated transcript model
        if len(molitem.mappings_record) == 0:
            return
        # Compatible with one or more transcript models:
        else:
            # Check that there are not different possible genes ??
            if len(set(i.geneid for i in molitem.mappings_record.keys())) == 1:
                gene_check: Set[str] = set()
                
                has_onlyintron_model = 0
                has_only_span_exin_model = 1
                has_onlyintron_and_valid_model = 0
                has_valid_mixed_model = 0
                has_invalid_mixed_model = 0
                has_onlyexo_model = 0
                has_mixed_model = 0
                multi_gene = 0
                for transcript_model, segments_list in molitem.mappings_record.items():
                    gene_check.add(transcript_model.geneid)
                    if len(gene_check) > 1:
                        multi_gene = 1
                    has_introns = 0
                    has_exons = 0
                    has_exseg_with_spliced_flag = 0
                    has_validated_intron = 0
                    has_exin_intron_span = 0
                    has_non3prime = 0
                    for segment_match in segments_list:
                        if segment_match.maps_to_intron:
                            has_introns = 1
                            if segment_match.feature.is_validated:
                                has_validated_intron = 1
                                if segment_match.feature.end_overlaps_with_part_of(segment_match.segment):
                                    downstream_exon = segment_match.feature.get_downstream_exon()
                                    if downstream_exon.start_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                                if segment_match.feature.start_overlaps_with_part_of(segment_match.segment):
                                    upstream_exon = segment_match.feature.get_upstream_exon()
                                    if upstream_exon.end_overlaps_with_part_of(segment_match.segment):
                                        has_exin_intron_span = 1
                        elif segment_match.maps_to_exon:
                            has_exons = 1
                            if not segment_match.feature.is_last_3prime:
                                has_non3prime = 1
                            if segment_match.is_spliced:
                                has_exseg_with_spliced_flag = 1
                    if has_validated_intron and not has_exons:
                        has_onlyintron_and_valid_model = 1
                    if has_introns and not has_exons:
                        has_onlyintron_model = 1
                    if has_exons and not has_introns:
                        has_onlyexo_model = 1
                    if has_exons and has_introns and not has_validated_intron and not has_exin_intron_span:
                        has_invalid_mixed_model = 1
                        has_mixed_model = 1
                    if has_exons and has_introns and has_validated_intron and not has_exin_intron_span:
                        has_valid_mixed_model = 1
                        has_mixed_model = 1
                    if not has_exin_intron_span:
                        has_only_span_exin_model = 0
                        
                if multi_gene:
                    # Many genes are compatible with the observation, do not count
                    return
                else:
                    if not len(molitem.mappings_record):
                        # No gene is compatible with the observation, do not count
                        return
                    else:
                        if has_onlyexo_model and not has_onlyintron_model and not has_mixed_model:
                            # More common situation, normal exonic read, count as spliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_only_span_exin_model:
                            # All the compatible transcript models have spanning exon-intron boundaries, count unspliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                            else:
                                # Non-singleton in validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                        if has_onlyintron_model and not has_onlyintron_and_valid_model and not has_mixed_model and not has_onlyexo_model:
                            if len(segments_list) == 1:
                                # Singleton in non-validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                            else:
                                # Non-singleton in non-validated intron
                                gene_ix = geneid2ix[transcript_model.geneid]
                                unspliced[gene_ix, cell_bcidx] += 1
                                return
                        if has_invalid_mixed_model and not has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Not validated and mapping to exon and introns, happens rarely in 10X / irrelevant. Count anyways
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_valid_mixed_model and not has_onlyintron_model and not has_onlyexo_model and not has_only_span_exin_model:
                            # Validated and mapping to exon and introns, happens rarely in 10X. Count as unspliced.
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and not has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping, most common case! Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and not has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if not has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping. Very rare. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return


class SmartSeq2(Logic):
    """SmartSeq2 logic
    """

    def __init__(self) -> None:
        self.name = "SmartSeq2"

    @property
    def layers(self) -> List[str]:  # This should be overridden if a different set of layers is desired
        return ["spliced", "unspliced", "ambiguous", "spanning"]

    @property
    def stranded(self) -> bool:
        return False

    @property
    def perform_validation_markup(self) -> bool:
        return False

    @property
    def accept_discordant(self) -> bool:
        return False

    def count(self, molitem: vcy.Molitem, cell_bcidx: int, dict_layers_columns: Dict[str, np.ndarray], geneid2ix: Dict[str, int]) -> None:
        # NOTE This can be simplified qyuite a bit, without loss of acuracy!
        # The hits are not compatible with any annotated transcript model
        spliced = dict_layers_columns["spliced"]
        unspliced = dict_layers_columns["unspliced"]
        ambiguous = dict_layers_columns["ambiguous"]
        spanning = dict_layers_columns["spanning"]

        if len(molitem.mappings_record) == 0:
            return
        # Compatible with one or more transcript models:
        else:
            # Check that there are not different possible genes ??
            if len(set(i.geneid for i in molitem.mappings_record.keys())) == 1:
                gene_check: Set[str] = set()
                
                has_onlyintron_model = 0
                has_only_span_exin_model = 1
                has_onlyexo_model = 0
                has_mixed_model = 0
                multi_gene = 0
                for transcript_model, segments_list in molitem.mappings_record.items():
                    gene_check.add(transcript_model.geneid)
                    if len(gene_check) > 1:
                        multi_gene = 1
                    has_introns = 0
                    has_exons = 0
                    has_exseg_with_spliced_flag = 0
                    has_exin_intron_span = 0
                    for segment_match in segments_list:
                        if segment_match.maps_to_intron:
                            has_introns = 1
                            if segment_match.feature.end_overlaps_with_part_of(segment_match.segment):
                                downstream_exon = segment_match.feature.get_downstream_exon()
                                if downstream_exon.start_overlaps_with_part_of(segment_match.segment):
                                    has_exin_intron_span = 1
                            if segment_match.feature.start_overlaps_with_part_of(segment_match.segment):
                                upstream_exon = segment_match.feature.get_upstream_exon()
                                if upstream_exon.end_overlaps_with_part_of(segment_match.segment):
                                    has_exin_intron_span = 1
                        elif segment_match.maps_to_exon:
                            has_exons = 1
                            if segment_match.is_spliced:
                                has_exseg_with_spliced_flag = 1
                    if has_introns and not has_exons:
                        has_onlyintron_model = 1
                    if has_exons and not has_introns:
                        has_onlyexo_model = 1
                    if has_exons and has_introns and not has_exin_intron_span:
                        has_valid_mixed_model = 1
                        has_mixed_model = 1
                    if not has_exin_intron_span:
                        has_only_span_exin_model = 0
                        
                if multi_gene:
                    # Many genes are compatible with the observation, do not count
                    return
                else:
                    if not len(molitem.mappings_record):
                        # NOTE it does not happen for Smartseq2
                        # No gene is compatible with the observation, do not count
                        return
                    else:
                        if has_onlyexo_model and not has_onlyintron_model and not has_mixed_model:
                            # More common situation, normal exonic read, count as spliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_only_span_exin_model:
                            # NOTE This is what I want to count as spanning
                            # All the compatible transcript models have spanning exon-intron boundaries, count unspliced
                            gene_ix = geneid2ix[transcript_model.geneid]
                            spanning[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and not has_mixed_model and not has_onlyexo_model:
                            gene_ix = geneid2ix[transcript_model.geneid]
                            unspliced[gene_ix, cell_bcidx] += 1
                            return
                        if has_onlyintron_model and has_onlyexo_model and not has_mixed_model:
                            # Ambiguity among the transcript models compatible with the mapping, most common case! Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return
                        if not has_onlyintron_model and has_onlyexo_model and has_mixed_model:
                            # NOTE has_mixed model is used only here in this logic
                            # Ambiguity among the transcript models compatible with the mapping. Count ambiguous
                            gene_ix = geneid2ix[transcript_model.geneid]
                            ambiguous[gene_ix, cell_bcidx] += 1
                            return


Default: type = Permissive10X



import numpy as np
import scipy.optimize
from scipy import sparse
import logging
from typing import *
from sklearn.neighbors import NearestNeighbors
from .speedboosted import _colDeltaCor, _colDeltaCorLog10, _colDeltaCorSqrt
from .speedboosted import _colDeltaCorpartial, _colDeltaCorLog10partial, _colDeltaCorSqrtpartial


def colDeltaCor(emat: np.ndarray, dmat: np.ndarray, threads: int=None) -> np.ndarray:
    """Calculate the correlation between the displacement (d[:,i])
    and the difference between a cell and every other (e - e[:, i])
    
    Parallel cython+OpenMP implemetation

    Arguments
    ---------
    emat: np.ndarray (ngenes, ncells)
        gene expression matrix
    dmat: np.ndarray (ngenes, ncells)
        gene velocity/displacement matrix
    threads: int
        number of parallel threads to use
    """
    import multiprocessing
    if threads is None:
        num_threads = int(multiprocessing.cpu_count() / 2)
    else:
        num_threads = max(threads, multiprocessing.cpu_count())
    out = np.zeros((emat.shape[1], emat.shape[1]))
    _colDeltaCor(emat, dmat, out, num_threads)
    return out


def colDeltaCorpartial(emat: np.ndarray, dmat: np.ndarray, ixs: np.ndarray, threads: int=None) -> np.ndarray:
    """Calculate the correlation between the displacement (d[:,i])
    and the difference between a cell and every other (e - e[:, i])
    
    Parallel cython+OpenMP implemetation

    Arguments
    ---------
    emat: np.ndarray (ngenes, ncells)
        gene expression matrix
    dmat: np.ndarray (ngenes, ncells)
        gene velocity/displacement matrix
    ixs: the neighborhood matrix (ncells, nneighbours)
        ixs[i, k] is the kth neighbour to the cell i
    threads: int
        number of parallel threads to use
    """
    import multiprocessing
    if threads is None:
        num_threads = int(multiprocessing.cpu_count() / 2)
    else:
        num_threads = max(threads, multiprocessing.cpu_count())
    out = np.zeros((emat.shape[1], emat.shape[1]))
    emat = np.require(emat, requirements="C")
    ixs = np.require(ixs, requirements="C").astype(np.intp)
    _colDeltaCorpartial(emat, dmat, out, ixs, num_threads)
    return out


def colDeltaCorLog10(emat: np.ndarray, dmat: np.ndarray, threads: int=None, psc: float=1.0) -> np.ndarray:
    """Calculate the correlation between the displacement (d[:,i])
    and the difference between a cell and every other (e - e[:, i])
    
    Parallel cython+OpenMP implemetation

    Arguments
    ---------
    emat: np.ndarray (ngenes, ncells)
        gene expression matrix
    dmat: np.ndarray (ngenes, ncells)
        gene velocity/displacement matrix
    threads: int
        number of parallel threads to use
    """
    import multiprocessing
    if threads is None:
        num_threads = int(multiprocessing.cpu_count() / 2)
    else:
        num_threads = max(threads, multiprocessing.cpu_count())
    out = np.zeros((emat.shape[1], emat.shape[1]))
    _colDeltaCorLog10(emat, dmat, out, num_threads, psc)
    return out


def colDeltaCorLog10partial(emat: np.ndarray, dmat: np.ndarray, ixs: np.ndarray, threads: int=None, psc: float=1.0) -> np.ndarray:
    """Calculate the correlation between the displacement (d[:,i])
    and the difference between a cell and every other (e - e[:, i])
    
    Parallel cython+OpenMP implemetation

    Arguments
    ---------
    emat: np.ndarray (ngenes, ncells)
        gene expression matrix
    dmat: np.ndarray (ngenes, ncells)
        gene velocity/displacement matrix
    ixs: the neighborhood matrix (ncells, nneighbours)
        ixs[i, k] is the kth neighbour to the cell i
    threads: int
        number of parallel threads to use
    """
    import multiprocessing
    if threads is None:
        num_threads = int(multiprocessing.cpu_count() / 2)
    else:
        num_threads = max(threads, multiprocessing.cpu_count())
    out = np.zeros((emat.shape[1], emat.shape[1]))
    emat = np.require(emat, requirements="C")
    ixs = np.require(ixs, requirements="C").astype(np.intp)
    _colDeltaCorLog10partial(emat, dmat, out, ixs, num_threads, psc)
    return out


def colDeltaCorSqrt(emat: np.ndarray, dmat: np.ndarray, threads: int=None, psc: float=0.0) -> np.ndarray:
    """Calculate the correlation between the displacement (d[:,i])
    and the difference between a cell and every other (e - e[:, i])
    
    Parallel cython+OpenMP implemetation

    Arguments
    ---------
    emat: np.ndarray (ngenes, ncells)
        gene expression matrix
    dmat: np.ndarray (ngenes, ncells)
        gene velocity/displacement matrix
    threads: int
        number of parallel threads to use
    """
    import multiprocessing
    if threads is None:
        num_threads = int(multiprocessing.cpu_count() / 2)
    else:
        num_threads = max(threads, multiprocessing.cpu_count())
    out = np.zeros((emat.shape[1], emat.shape[1]))
    _colDeltaCorSqrt(emat, dmat, out, num_threads, psc)
    return out


def colDeltaCorSqrtpartial(emat: np.ndarray, dmat: np.ndarray, ixs: np.ndarray, threads: int=None, psc: float=0.0) -> np.ndarray:
    """Calculate the correlation between the displacement (d[:,i])
    and the difference between a cell and every other (e - e[:, i])
    
    Parallel cython+OpenMP implemetation

    Arguments
    ---------
    emat: np.ndarray (ngenes, ncells)
        gene expression matrix
    dmat: np.ndarray (ngenes, ncells)
        gene velocity/displacement matrix
    ixs: the neighborhood matrix (ncells, nneighbours)
        ixs[i, k] is the kth neighbour to the cell i
    threads: int
        number of parallel threads to use
    """
    import multiprocessing
    if threads is None:
        num_threads = int(multiprocessing.cpu_count() / 2)
    else:
        num_threads = max(threads, multiprocessing.cpu_count())
    out = np.zeros((emat.shape[1], emat.shape[1]))
    emat = np.require(emat, requirements="C")
    ixs = np.require(ixs, requirements="C").astype(np.intp)
    _colDeltaCorSqrtpartial(emat, dmat, out, ixs, num_threads, psc)
    return out


def _fit1_slope(y: np.ndarray, x: np.ndarray) -> float:
    """Simple function that fit a linear regression model without intercept
    """
    if not np.any(x):
        m = np.NAN  # It is definetelly not at steady state!!!
    elif not np.any(y):
        m = 0
    else:
        result, rnorm = scipy.optimize.nnls(x[:, None], y)  # Fastest but costrains result >= 0
        m = result[0]
        # Second fastest: m, _ = scipy.optimize.leastsq(lambda m: x*m - y, x0=(0,))
        # Third fastest: m = scipy.optimize.minimize_scalar(lambda m: np.sum((x*m - y)**2 )).x
        # Before I was doinf fastest: scipy.optimize.minimize_scalar(lambda m: np.sum((y - m * x)**2), bounds=(0, 3), method="bounded").x
        # Optionally one could clip m if high value make no sense
        # m = np.clip(m,0,3)
    return m


def _fit1_slope_weighted(y: np.ndarray, x: np.ndarray, w: np.ndarray, limit_gamma: bool=False, bounds: Tuple[float, float]=(0, 20)) -> float:
    """Simple function that fit a weighted linear regression model without intercept
    """
    if not np.any(x):
        m = np.NAN  # It is definetelly not at steady state!!!
    elif not np.any(y):
        m = 0
    else:
        if limit_gamma:
            if np.median(y) > np.median(x):
                high_x = x > np.percentile(x, 90)
                up_gamma = np.percentile(y[high_x], 10) / np.median(x[high_x])
                up_gamma = np.maximum(1.5, up_gamma)
            else:
                up_gamma = 1.5  # Just a bit more than 1
            m = scipy.optimize.minimize_scalar(lambda m: np.sum(w * (x * m - y)**2), bounds=(1e-8, up_gamma), method="bounded").x
        else:
            m = scipy.optimize.minimize_scalar(lambda m: np.sum(w * (x * m - y)**2), bounds=bounds, method="bounded").x
    return m


def _fit1_slope_weighted_offset(y: np.ndarray, x: np.ndarray, w: np.ndarray, fixperc_q: bool=False, limit_gamma: bool=False) -> Tuple[float, float]:
    """Function that fits a weighted linear regression model with intercept
    with some adhoc
    """
    if not np.any(x):
        m = (np.NAN, 0)  # It is definetelly not at steady state!!!
    elif not np.any(y):
        m = (0, 0)
    else:
        if fixperc_q:
            m1 = np.percentile(y[x <= np.percentile(x, 1)], 50)
            m0 = scipy.optimize.minimize_scalar(lambda m: np.sum(w * (x * m - y + m1)**2), bounds=(0, 20), method="bounded").x
            m = (m0, m1)
        else:
            # m, _ = scipy.optimize.leastsq(lambda m: np.sqrt(w) * (-y + x * m[0] + m[1]), x0=(0, 0))  # This is probably faster but it can have negative slope
            # NOTE: The up_gamma is to deal with cases where consistently y > x. Those should have positive velocity everywhere
            if limit_gamma:
                if np.median(y) > np.median(x):
                    high_x = x > np.percentile(x, 90)
                    up_gamma = np.percentile(y[high_x], 10) / np.median(x[high_x])
                    up_gamma = np.maximum(1.5, up_gamma)
                else:
                    up_gamma = 1.5  # Just a bit more than 1
            else:
                up_gamma = 20
            up_q = 2 * np.sum(y * w) / np.sum(w)
            m = scipy.optimize.minimize(lambda m: np.sum(w * (-y + x * m[0] + m[1])**2),
                                        x0=(0.1, 1e-16), method="L-BFGS-B",
                                        bounds=[(1e-8, up_gamma), (0, up_q)]).x  # If speedup is needed either the gradient or numexpr could be used
    return m[0], m[1]


def _fit1_slope_offset(y: np.ndarray, x: np.ndarray, fixperc_q: bool=False) -> Tuple[float, float]:
    """Simple function that fit a linear regression model with intercept
    """
    if not np.any(x):
        m = (np.NAN, 0)  # It is definetelly not at steady state!!!
    elif not np.any(y):
        m = (0, 0)
    else:
        # result, rnorm = scipy.optimize.nnls(x[:, None], y)  # Fastest but costrains result >= 0
        # m = result[0]
        if fixperc_q:
            m1 = np.percentile(y[x <= np.percentile(x, 1)], 50)
            m0 = scipy.optimize.minimize_scalar(lambda m: np.sum((x * m - y + m1)**2), bounds=(0, 20), method="bounded").x
            m = (m0, m1)
        else:
            m, _ = scipy.optimize.leastsq(lambda m: -y + x * m[0] + m[1], x0=(0, 0))
        # Third fastest: m = scipy.optimize.minimize_scalar(lambda m: np.sum((x*m - y)**2 )).x
        # Before I was doinf fastest: scipy.optimize.minimize_scalar(lambda m: np.sum((y - m * x)**2), bounds=(0, 3), method="bounded").x
        # Optionally one could clip m if high value make no sense
        # m = np.clip(m,0,3)
    return m[0], m[1]


def fit_slope(Y: np.ndarray, X: np.ndarray) -> np.ndarray:
    """Loop through the genes and fits the slope

    Y: np.ndarray, shape=(genes, cells)
        the dependent variable (unspliced)
    X: np.ndarray, shape=(genes, cells)
        the independent variable (spliced)
    """
    # NOTE this could be easily parallelized
    slopes = np.fromiter((_fit1_slope(Y[i, :], X[i, :]) for i in range(Y.shape[0])),
                         dtype="float32",
                         count=Y.shape[0])
    return slopes


def fit_slope_offset(Y: np.ndarray, X: np.ndarray, fixperc_q: bool=False) -> Tuple[np.ndarray, np.ndarray]:
    """Loop through the genes and fits the slope

    Y: np.ndarray, shape=(genes, cells)
        the dependent variable (unspliced)
    X: np.ndarray, shape=(genes, cells)
        the independent variable (spliced)
    """
    # NOTE this could be easily parallelized
    slopes = np.zeros(Y.shape[0], dtype="float32")
    offsets = np.zeros(Y.shape[0], dtype="float32")
    for i in range(Y.shape[0]):
        m, q = _fit1_slope_offset(Y[i, :], X[i, :], fixperc_q)
        slopes[i] = m
        offsets[i] = q
    return slopes, offsets


def fit_slope_weighted(Y: np.ndarray, X: np.ndarray, W: np.ndarray, return_R2: bool=False, limit_gamma: bool=False, bounds: Tuple[float, float]=(0, 20)) -> np.ndarray:
    """Loop through the genes and fits the slope

    Y: np.ndarray, shape=(genes, cells)
        the dependent variable (unspliced)
    X: np.ndarray, shape=(genes, cells)
        the independent variable (spliced)
    W: np.ndarray, shape=(genes, cells)
        the weights that will scale the square residuals
    """
    # NOTE this could be easily parallelized
    # slopes = np.fromiter((_fit1_slope_weighted(Y[i, :], X[i, :], W[i, :], bounds=bounds) for i in range(Y.shape[0])),
    #                     dtype="float32",
    #                     count=Y.shape[0])

    slopes = np.zeros(Y.shape[0], dtype="float32")
    offsets = np.zeros(Y.shape[0], dtype="float32")
    if return_R2:
        R2 = np.zeros(Y.shape[0], dtype="float32")
    for i in range(Y.shape[0]):
        m = _fit1_slope_weighted(Y[i, :], X[i, :], W[i, :], limit_gamma)
        slopes[i] = m
        if return_R2:
            # NOTE: the coefficient of determination is not weighted but the fit is
            with np.errstate(divide='ignore', invalid='ignore'):
                SSres = np.sum((m * X[i, :] - Y[i, :])**2)
                SStot = np.sum((Y[i, :].mean() - Y[i, :])**2)
                _R2 = 1 - (SSres / SStot)
            if np.isfinite(_R2):
                R2[i] = _R2
            else:
                R2[i] = -1e16
    if return_R2:
        return slopes, R2
    return slopes


def fit_slope_weighted_offset(Y: np.ndarray, X: np.ndarray, W: np.ndarray, fixperc_q: bool=False, return_R2: bool=True, limit_gamma: bool=False) -> Any:
    """Loop through the genes and fits the slope

    Y: np.ndarray, shape=(genes, cells)
        the dependent variable (unspliced)
    X: np.ndarray, shape=(genes, cells)
        the independent variable (spliced)
    """
    # NOTE this could be easily parallelized
    slopes = np.zeros(Y.shape[0], dtype="float32")
    offsets = np.zeros(Y.shape[0], dtype="float32")
    if return_R2:
        R2 = np.zeros(Y.shape[0], dtype="float32")
    for i in range(Y.shape[0]):
        m, q = _fit1_slope_weighted_offset(Y[i, :], X[i, :], W[i, :], fixperc_q, limit_gamma)
        slopes[i] = m
        offsets[i] = q
        if return_R2:
            # NOTE: the coefficient of determination is not weighted but the fit is
            with np.errstate(divide='ignore', invalid='ignore'):
                SSres = np.sum((m * X[i, :] + q - Y[i, :])**2)
                SStot = np.sum((Y[i, :].mean() - Y[i, :])**2)
                _R2 = 1 - (SSres / SStot)
            if np.isfinite(_R2):
                R2[i] = _R2
            else:
                R2[i] = -1e16
    if return_R2:
        return slopes, offsets, R2
    return slopes, offsets


def clusters_stats(U: np.ndarray, S: np.ndarray,
                   clusters_uid: np.ndarray, cluster_ix: np.ndarray,
                   size_limit: int=40) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate the averages per cluster
    
    If the cluster is too small (size<size_limit) the average of the toal is reported instead
    """
    U_avgs = np.zeros((S.shape[0], len(clusters_uid)))
    S_avgs = np.zeros((S.shape[0], len(clusters_uid)))
    avgU_div_avgS = np.zeros((S.shape[0], len(clusters_uid)))
    slopes_by_clust = np.zeros((S.shape[0], len(clusters_uid)))
    for i, uid in enumerate(clusters_uid):
        cluster_filter = cluster_ix == i
        n_cells = np.sum(cluster_filter)
        logging.info(f"Cluster: {uid} ({n_cells} cells)")
        if n_cells > size_limit:
            U_avgs[:, i], S_avgs[:, i] = U[:, cluster_filter].mean(1), S[:, cluster_filter].mean(1)
        else:
            U_avgs[:, i], S_avgs[:, i] = U.mean(1), S.mean(1)
            
    return U_avgs, S_avgs



import sys
import os
import glob
import re
import click
import numpy as np
import random
import string
import logging
from typing import *
import velocyto as vcy
from ._run import _run

# logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)


def id_generator(size: int=6, chars: str=string.ascii_uppercase + string.digits) -> str:
    return ''.join(random.choice(chars) for _ in range(size))


@click.command(short_help="Runs the velocity analysis outputing a loom file")
@click.argument("bamfile", nargs=-1, required=True,
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.argument("gtffile",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.option("--bcfile", "-b",
              help="""Valid barcodes file, to filter the bam. If --bcfile is not specified all the cell barcodes will be incuded.
              Cell barcodes should be specified in the bcfile as the `CB` tag for each read""",
              default=None,
              show_default=True,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--outputfolder", "-o",
              help="Output folder, if it does not exist it will be created.",
              default=None,
              type=click.Path(exists=False))
@click.option("--sampleid", "-e",
              help="The sample name that will be used to retrieve informations from metadatatable",
              default=None,
              type=click.Path(exists=False))
@click.option("--metadatatable", "-s",
              help="Table containing metadata of the various samples (csv formatted, rows are samples and cols are entries)",
              default=None,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--mask", "-m",
              help=".gtf file containing intervals to mask",
              default=None,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--onefilepercell", "-c",
              help="""If this flag is used every bamfile passed is interpreted as an independent cell, otherwise multiple files are interpreted as batch of different cells to be analized together.
              Important: cells reads should not be distributed over multiple bamfiles is not supported!! (default: off)""",
              default=False,
              is_flag=True)
@click.option("--logic", "-l",
              help="The logic to use for the filtering (default: Default)",
              default="Default")
@click.option("--without-umi", "-U",
              help="If this flag is used the data is assumed UMI-less and reads are counted instead of molecules (default: off)",
              default=False,
              is_flag=True)
@click.option("--umi-extension", "-u",
              help="""In case UMI is too short to guarantee uniqueness (without information from the ampping) set this parameter to `chr`, `Gene` ro `[N]bp`
              If set to `chr` the mapping position (binned to 10Gb intervals) will be appended to `UB` (ideal for InDrops+dropEst). If set to `Gene` then the `GX` tag will be appended to the `UB` tag.
              If set to `[N]bp` the first N bases of the sequence will be used to extend `UB` (ideal for STRT). (Default: `no`)""",
              default="no")
@click.option("--multimap", "-M",
              help="""Consider not unique mappings (not reccomended)""",
              default=False,
              is_flag=True)
@click.option("--samtools-threads", "-@",
              help="The number of threads to use to sort the bam by cellID file using samtools",
              default=16)
@click.option("--samtools-memory",
              help="The number of MB used for every thread by samtools to sort the bam file",
              default=2048)
@click.option("--dump", "-d",
              help="For debugging purposes only: it will dump a molecular mapping report to hdf5. --dump N, saves a cell every N cells. If p is prepended a more complete (but huge) pickle report is printed (default: 0)",
              default="0")
@click.option('--verbose', '-v',
              help="Set the vebosity level: -v (only warinings) -vv (warinings and info) -vvv (warinings, info and debug)",
              count=True, default=1)
def run(bamfile: str, gtffile: str,
        bcfile: str, outputfolder: str,
        sampleid: str, metadatatable: str,
        mask: str, onefilepercell: bool, logic: str, without_umi: str, umi_extension: str, multimap: bool,
        samtools_threads: int, samtools_memory: int, dump: str, verbose: int,
        additional_ca: dict={}) -> None:
    """Runs the velocity analysis outputing a loom file

    BAMFILE bam file with sorted reads

    GTFFILE genome annotation file
    """
    return _run(bamfile=bamfile, gtffile=gtffile, bcfile=bcfile, outputfolder=outputfolder,
                sampleid=sampleid, metadatatable=metadatatable, repmask=mask, onefilepercell=onefilepercell,
                logic=logic, without_umi=without_umi, umi_extension=umi_extension, multimap=multimap, test=False, samtools_threads=samtools_threads,
                samtools_memory=samtools_memory, dump=dump, verbose=verbose, additional_ca=additional_ca)



import sys
import os
import glob
import re
import gzip
import click
import loompy
import numpy as np
import random
import string
import csv
from collections import defaultdict
import logging
from typing import *
import velocyto as vcy
from ._run import _run

# logging.basicConfig(stream=sys.stdout, format='%(asctime)s - %(levelname)s - %(message)s', level=logging.DEBUG)


def id_generator(size: int=6, chars: str=string.ascii_uppercase + string.digits) -> str:
    return ''.join(random.choice(chars) for _ in range(size))


@click.command(short_help="Runs the velocity analysis on SmartSeq2 data (independent bam file per cell)")
@click.argument("bamfiles", nargs=-1, required=True,
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.argument("gtffile",
                type=click.Path(exists=True,
                                file_okay=True,
                                dir_okay=False,
                                readable=True,
                                resolve_path=True))
@click.option("--outputfolder", "-o",
              help="Output folder, if it does not exist it will be created.",
              default=None,
              type=click.Path(exists=False))
@click.option("--sampleid", "-e",
              help="The sample name that will be used as a the filename of the output.",
              default=None,
              type=click.Path(exists=False))
@click.option("--repmask", "-m",
              help=".gtf file containing intervals to mask",
              default=None,
              type=click.Path(resolve_path=True,
                              file_okay=True,
                              dir_okay=False,
                              readable=True))
@click.option("--dump", "-d",
              help="For debugging purposes only: it will dump a molecular mapping report to hdf5. --dump N, saves a cell every N cells. If p is prepended a more complete (but huge) pickle report is printed (default: 0)",
              default="0")
@click.option('--verbose', '-v',
              help="Set the vebosity level: -v (only warinings) -vv (warinings and info) -vvv (warinings, info and debug)",
              count=True, default=1)
def run_smartseq2(bamfiles: str, gtffile: str, outputfolder: str, sampleid: str,
                  repmask: str, dump: str, verbose: int, additional_ca: dict={}) -> None:
    """Runs the velocity analysis on SmartSeq2 data (independent bam file per cell)

    [BAMFILES, ...] a sequence of bam files to be analyzed (e.g. use a wild-card expansion)

    GTFFILE genome annotation file
    """
    return _run(bamfile=bamfiles, gtffile=gtffile, bcfile=None, outputfolder=outputfolder,
                sampleid=sampleid, metadatatable=None, repmask=repmask, onefilepercell=True,
                logic="SmartSeq2", without_umi=True, umi_extension="no",
                multimap=False, test=False, samtools_threads=1,
                samtools_memory=1, dump=dump, verbose=verbose, additional_ca=additional_ca)
